#include <iostream>
#include <cstring>
#include <string>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <limits>
#include <algorithm>
#include <sys/time.h>
#include <sys/stat.h>
#include <getopt.h> 
#include <cstdlib>
#include <gsl/gsl_sf_gamma.h>

#include "src/AlphaStrong.hh"
#include "src/ObservableScaleDep.hh"
#include "src/THunc_model.hh"
#include "src/ScaleIndependentResult.hh"
#include "src/StudyDistribution.hh"

using namespace std;



// insert here the information on the perturbative expansion you want to consider

double *initProcess(string &process, int &nOrders, double &Q, double &muR, double &as_muR, int &LO_as_power, double &muRmin, double &muRmax, int &Np, double &dmin, double &dmax)
{
  // ggH @ 13TeV with muF=mH/2, PDF4LHC15_nnlo_100, rEFT, mt=172.5, no bottom and no charm    ggHiggs -f 0.5
  process = "ggH";  // name (to be used in output files)
  nOrders = 4;      // number of known orders
  double *Sigma = new double[nOrders];
  Sigma[0] = 12.9953;  // LO
  Sigma[1] = 30.705;   // NLO
  Sigma[2] = 41.8182;  // NNLO
  Sigma[3] = 46.2879;  // N3LO  and so on
  //
  Q   = 125.;       // hard scale of the process
  muR = Q;          // renormalization scale at which the expansion is computed
  as_muR = 0.112638;  // value of the stron coupling at muR
  LO_as_power = 2;    // the power of alpha_s of the LO of the process
  // the range of the renormaization scale used as support of the prior for computing scale independent results
  muRmin = Q/10.;   // lower limit
  muRmax = Q*2;     // upper limit
  //
  Np = 8234;   // the number of points used to sample the distribution in the most important region, defined by the range below
  dmin = 0;    // lower limit of the interesting range in units of the LO
  dmax = 4;    // upper limit of the interesting range in units of the LO
  //
  return Sigma;
}





// The following are three examples (discussed in the paper) of possible models implemented through the "custom model" feature
// To activate one of them, uncomment it



/*
// scale variation model with violation of the bound
#define CustomModel
const int customNpars = 1;
const bool isNormalized = true;
double custom_P_deltan_given_pars(double *delta, double *rhat, int k, const double *par) {
  double lam = par[0]/(1-par[0]);
  double rho = 2*k*k;
  return exp(rho*(1-max(1.,fabs(delta[k])/lam/rhat[k-1])))/(1+1./rho) /lam/2./rhat[k-1];
}
double custom_P0(const double *par) {
  double gamma = 1;
  double lambda = par[0]/(1-par[0]);
  double jac = 1./(1-par[0])/(1-par[0]);
  return pow(lambda,gamma)*exp(-lambda) /gsl_sf_gamma(1+gamma) * jac;
}
//*/



/*
// scale variation model with controlled scale dependence (i.e., with constraint on r_k)
#define CustomModel
const int customNpars = 2;
const bool isNormalized = false;
double custom_P_deltan_given_pars(double *delta, double *rhat, int k, const double *par) {
  double lam = par[0]/(1-par[0]);
  double eta = par[1]/(1-par[1]);
  double er  = eta*rhat[k-1];
  if(rhat[k] > er) return 0;
  double den = rhat[k-1]*lam;
  if(fabs(delta[k]) > den) return 0;
  return 1./2/den;
}
double custom_P0(const double *par) {
  double gamma = 1;
  double lam = par[0]/(1-par[0]);
  double jac = 1./(1-par[0])/(1-par[0]);
  double P0 = pow(lam,gamma)*exp(-lam) /gsl_sf_gamma(1+gamma);
  double eta = par[1]/(1-par[1]);
  jac *= 1./(1-par[1])/(1-par[1]);
  P0 *= exp(-eta);
  return P0 * jac;
}
//*/



/*
// scale variation model + geometric behaviour model
#define CustomModel
const int customNpars = 3;
const bool isNormalized = true;
double custom_P_deltan_given_pars(double *delta, double *rhat, int k, const double *par) {
  double lam = par[0]/(1-par[0]);
  double c   = 1+ par[1]/(1-par[1]);
  double a   = par[2];
  double den = min(rhat[k-1]*lam, c*pow(a,k));
  if(fabs(delta[k]) > den) return 0;
  return 1./2/den;
}
double custom_P0(const double *par) {
  double gamma = 1;
  double lambda = par[0]/(1-par[0]);
  double c   = 1+ par[1]/(1-par[1]);
  double a   = par[2];
  double jac = 1./(1-par[0])/(1-par[0])/(1-par[1])/(1-par[1]);
  double P0 = pow(lambda,gamma)*exp(-lambda) /gsl_sf_gamma(1+gamma);
  double feps = 0.1;
  P0 *= feps/pow(c,1+feps) * 2*(1-a);
  return P0 * jac;
}
//*/






// if no custom model is given
#ifndef CustomModel
double custom_P_deltan_given_pars(double *delta, double *rhat, int k, const double *par) {
  return 0;
}
double custom_P0(const double *par) {
  return 0;
}
#endif






void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -r  --muR <muR/mH>        set muR/mH" << endl
       << " -s  --scaleIndep          compute just scale independent result" << endl
       << " -P  --PHp <PHp>           set the prior of the hypotesys H+,H- (sign patterns)" << endl
       << endl;
  exit(0);
  return;
}

void read_arguments(int argc, char* argv[], double &muRrat, bool &doScaleIndep, double &probHplus) {
  const char* const short_options = "hr:sP:";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "muR", 1, NULL, 'r' },
					 { "scaleIndep", 0, NULL, 's' },
					 { "PHp", 1, NULL, 'P' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 's':
      doScaleIndep = true;
      break;
    case 'r':
      muRrat = strtod(optarg, NULL);
      break;
    case 'P':
      probHplus = strtod(optarg, NULL);
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  //
  return;
}


double delta_t(struct timeval t0, struct timeval t1) {
  return t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
}



//
//  Main program
//
int main(int argc, char* argv[])
{
  
  struct timeval t0, t, tlast;
  gettimeofday(&t0,NULL);
  tlast = t0;

  bool doScaleIndep = false;
  double muRrat = 1;
  double probHplus = 0;
  read_arguments(argc, argv, muRrat, doScaleIndep, probHplus);


  int nOrd;
  double mH, muR0, asR0;
  double muRmin, muRmax;
  double Procdmin=0, Procdmax=2;
  int offset, ProcNp=0;
  string proc;
  double *isigma = initProcess(proc, nOrd, mH, muR0, asR0, offset, muRmin, muRmax, ProcNp, Procdmin, Procdmax);



  // used to plot the scale dependence
  QCDobservable *Obs = new QCDobservable(isigma, nOrd, offset, muR0, asR0, 5);

  double *osigma, *r;
  osigma = new double[nOrd];
  r      = new double[nOrd];


  const int Npoints = 331;
  double muR, muRminPlot = max(mH/pow(2.,6),1.61), muRmaxPlot = mH*pow(2.,7);
  //
  ofstream ofile1("output/"+proc+"_scale_dependence.dat");
  for(int k=Npoints-1; k>=0; k--) {
    muR = muRminPlot * exp(k/(Npoints-1.)*log(muRmaxPlot/muRminPlot));
    ofile1 << setw(12) << muR/mH;
    Obs->eval(muR,osigma);
    for(int i=0; i<nOrd; i++) ofile1 << setw(14) << osigma[i];
    ofile1 << endl;
  }
  ofile1.close();

  gettimeofday(&t,NULL);
  cout << "Step 1 done in " << delta_t(tlast,t) << " s" << endl;
  tlast = t;





  // compute the XS at the central scale 
  Obs->eval(muRrat*mH,osigma,r);


  // canonical scale variation uncertainty
  double *CSU_up   = new double[nOrd];
  double *CSU_down = new double[nOrd];
  Obs->ConventionalScaleUncertainty(muRrat*mH, CSU_up, CSU_down, 2, 5);





  ostringstream muRs;
  muRs << muRrat;
  ostringstream PHpms;
  if(probHplus>0) PHpms << "_PHpm" << probHplus;


  int Np = 8800;
  double dmin=-60, dmax=62;
  vector<double> xSigma, xSigmaLess;
  for(int k=0; k<Np; k++) {
    double d = dmin + k/(Np-1.)*(dmax-dmin);
    xSigma.push_back(d*osigma[0]);
    if(k%100==0 && d>-3 && d<6) xSigmaLess.push_back(d*osigma[0]);
  }
  //
  for(int k=0; k<ProcNp; k++) {
    double d = Procdmin + k/(ProcNp-1.)*(Procdmax-Procdmin);
    xSigma.push_back(d*osigma[0]);
    if(k%20==0 && d>-3 && d<6) xSigmaLess.push_back(d*osigma[0]);
  }
  sort(xSigma.begin(), xSigma.end());
  sort(xSigmaLess.begin(), xSigmaLess.end());
  Np = xSigma.size();
  //


  double *valSigma = new double[Np];
#ifdef CustomModel
  double *valSigmaLess = new double[xSigmaLess.size()];
#endif
  map<string,double*> valPSigmanp1[10];
  map<string,StudyDistribution*> PSigmanp1[10];
  map<string,double> stdev[10];
  map<string,double> xplus68[10], xminus68[10];
  map<string,double> xplus95[10], xminus95[10];
  //map<string,double> Pneg[10];
  for(int n=0; n<nOrd; n++) {
    valPSigmanp1[n]["Geo"] = new double[Np];
    valPSigmanp1[n]["Sca"] = new double[Np];
    valPSigmanp1[n]["Cus"] = new double[xSigmaLess.size()];
  }
  double stdevGeo[10];
  ofstream ofile2, ofile2sum;
  string sorder[] = { "LO", "NLO", "NNLO", "N3LO", "N4LO", "N5LO", "N6LO" };



  gettimeofday(&t,NULL);
  cout << "Step 2 done in " << delta_t(tlast,t) << " s" << endl;
  tlast = t;
  //



  if(doScaleIndep) {  // scale independent case
    ofile2   .open("output/"+proc+"_P_Sigma_np1_muRindep"+PHpms.str()+".dat");
    ofile2sum.open("output/"+proc+"_summary"+"_muRindep"+PHpms.str()+".dat");
    //
    // geometric behaviour model
    ScaleIndependentResult<THuncModel_Geo> *SIresG = new ScaleIndependentResult<THuncModel_Geo>(Obs);
    SIresG->SetPHplusminus(probHplus);
    SIresG->SetMuPrior(flat,muRmin,muRmax);
    //
    // scale variation model
    ScaleIndependentResult<THuncModel_Scale> *SIresS = new ScaleIndependentResult<THuncModel_Scale>(Obs);
    SIresS->SetPHplusminus(probHplus);
    SIresS->SetMuPrior(flat,muRmin,muRmax);
    //
#ifdef CustomModel
    // custom model
    ScaleIndependentResult<THuncModel_Custom> *SIresC = new ScaleIndependentResult<THuncModel_Custom>(Obs);
    SIresC->SetPHplusminus(probHplus);
    SIresC->SetMuPrior(flat,muRmin,muRmax);
#endif
    //
    for(int n=0; n<nOrd; n++) {
      double ESsqGeo = SIresG->E_Sigmanp1Sq_symm(n);
      double ESGeosq = pow(SIresG->E_Sigmanp1(n),2);
      stdevGeo[n] = sqrt(ESsqGeo-ESGeosq);
      //if(n==0 || n==1) stdevGeo[n] = __builtin_nan(""); // it depends on the prior for c
      if(stdevGeo[n]*stdevGeo[n]<0.02*max(ESsqGeo,ESGeosq)) stdevGeo[n] = -1; // when the std dev is the tiny difference of two large numbers, use the computation from the distribution below
      for(int k=0; k<Np; k++) {
	valPSigmanp1[n]["Geo"][k] = SIresG->P_Sigmanp1(xSigma[k],n);
	valPSigmanp1[n]["Sca"][k] = SIresS->P_Sigmanp1(xSigma[k],n);
	if(n==0) valSigma[k] = xSigma[k];
      }
#ifdef CustomModel
      for(int k=0; k<xSigmaLess.size(); k++) {
	valPSigmanp1[n]["Cus"][k] = SIresC->P_Sigmanp1(xSigmaLess[k],n);
	//cout << valPSigmanp1[n]["Cus"][k] << endl;
	if(n==0) valSigmaLess[k] = xSigmaLess[k];
	if(k%10==0) cout << "."; cout.flush();
      }
      cout << endl;
#endif
    }
  }
  //
  //
  else { // fixed scale case
    ofile2   .open("output/"+proc+"_P_Sigma_np1_muR"+muRs.str()+PHpms.str()+".dat");
    ofile2sum.open("output/"+proc+"_summary"+"_muR"+muRs.str()+PHpms.str()+".dat");
    //
    // geometric behaviour model
    THuncModel_Geo    *modelG = new THuncModel_Geo   (osigma,    nOrd, probHplus);
    //
    // scale variation model
    //THuncModel_Scale  *modelS = new THuncModel_Scale (osigma, r, nOrd, probHplus);
    THuncModel_Scale  *modelS = new THuncModel_Scale (Obs, muRrat*mH, probHplus);
    //
    // custom model
#ifdef CustomModel
    //THuncModel_Custom *modelC = new THuncModel_Custom(osigma, r, nOrd, probHplus);
    THuncModel_Custom *modelC = new THuncModel_Custom(Obs, muRrat*mH, probHplus);
    modelC->SetNumberOfParameters(customNpars);
#endif
    //
    for(int n=0; n<nOrd; n++) {
      stdevGeo[n] = osigma[0]*sqrt(modelG->Var_deltanp1(n));
      for(int k=0; k<Np; k++) {
	valPSigmanp1[n]["Geo"][k] = modelG->P_Sigmanp1(xSigma[k],n);
	valPSigmanp1[n]["Sca"][k] = modelS->P_Sigmanp1(xSigma[k],n);
	if(n==0) valSigma[k] = xSigma[k];
      }
#ifdef CustomModel
      for(int k=0; k<xSigmaLess.size(); k++) {
	valPSigmanp1[n]["Cus"][k] = 1;
	valPSigmanp1[n]["Cus"][k] = modelC->P_Sigmanp1(xSigmaLess[k],n);
	//cout << valPSigmanp1[n]["Cus"][k] << endl;
	if(n==0) valSigmaLess[k] = xSigmaLess[k];
	if(k%10==0) cout << "."; cout.flush();
      }
      cout << endl;
#endif
    }
  }


  gettimeofday(&t,NULL);
  cout << "Step 3 done in " << delta_t(tlast,t) << " s" << endl;
  tlast = t;



  for(int n=0; n<nOrd; n++) {
    // study the distributions computed above
    PSigmanp1[n]["Geo"] = new StudyDistribution(valSigma,    valPSigmanp1[n]["Geo"],Np);
    PSigmanp1[n]["Sca"] = new StudyDistribution(valSigma,    valPSigmanp1[n]["Sca"],Np);
#ifdef CustomModel
    PSigmanp1[n]["Cus"] = new StudyDistribution(valSigmaLess,valPSigmanp1[n]["Cus"],xSigmaLess.size());
    if(!isNormalized) PSigmanp1[n]["Cus"]->Normalize();
#endif
    // compute std deviation
    stdev[n]["Geo"] = stdevGeo[n];
    if(stdevGeo[n]<0) stdev[n]["Geo"] = sqrt(PSigmanp1[n]["Geo"]->Moment(2,PSigmanp1[n]["Geo"]->getMean()));
    stdev[n]["Sca"] = sqrt(PSigmanp1[n]["Sca"]->Moment(2,PSigmanp1[n]["Sca"]->getMean()));
#ifdef CustomModel
    stdev[n]["Cus"] = sqrt(PSigmanp1[n]["Cus"]->Moment(2,PSigmanp1[n]["Cus"]->getMean()));
#endif
    //
    // compute DoB intervals
    PSigmanp1[n]["Geo"]->DoBinterval(PSigmanp1[n]["Geo"]->getMedian(), 0.68, xplus68[n]["Geo"], xminus68[n]["Geo"]);
    PSigmanp1[n]["Geo"]->DoBinterval(PSigmanp1[n]["Geo"]->getMedian(), 0.95, xplus95[n]["Geo"], xminus95[n]["Geo"]);
    PSigmanp1[n]["Sca"]->DoBinterval(PSigmanp1[n]["Sca"]->getMedian(), 0.68, xplus68[n]["Sca"], xminus68[n]["Sca"]);
    PSigmanp1[n]["Sca"]->DoBinterval(PSigmanp1[n]["Sca"]->getMedian(), 0.95, xplus95[n]["Sca"], xminus95[n]["Sca"]);
#ifdef CustomModel
    PSigmanp1[n]["Cus"]->DoBinterval(PSigmanp1[n]["Cus"]->getMedian(), 0.68, xplus68[n]["Cus"], xminus68[n]["Cus"]);
    PSigmanp1[n]["Cus"]->DoBinterval(PSigmanp1[n]["Cus"]->getMedian(), 0.95, xplus95[n]["Cus"], xminus95[n]["Cus"]);
#endif
    // print results
    cout << "Order n=" << n << endl;
    cout << "           "
         << setw(11) << "Geo model"
         << setw(14) << "Scale model"
#ifdef CustomModel
         << setw(14) << "Custom model"
#endif
	 << endl;
    cout << "  Norm:    "
	 << setw(11) << PSigmanp1[n]["Geo"]->getNorm()
	 << setw(14) << PSigmanp1[n]["Sca"]->getNorm()
#ifdef CustomModel
	 << setw(14) << PSigmanp1[n]["Cus"]->getNorm()
#endif
	 << endl;
    cout << "  Mean:    "
	 << setw(11) << PSigmanp1[n]["Geo"]->getMean()
	 << setw(14) << PSigmanp1[n]["Sca"]->getMean()
#ifdef CustomModel
	 << setw(14) << PSigmanp1[n]["Cus"]->getMean()
#endif
	 << endl;
    cout << "  Mode:    "
	 << setw(11) << PSigmanp1[n]["Geo"]->getMode()
	 << setw(14) << PSigmanp1[n]["Sca"]->getMode()
#ifdef CustomModel
	 << setw(14) << PSigmanp1[n]["Cus"]->getMode()
#endif
	 << endl;
    cout << "  Median:  "
	 << setw(11) << PSigmanp1[n]["Geo"]->getMedian()
	 << setw(14) << PSigmanp1[n]["Sca"]->getMedian()
#ifdef CustomModel
	 << setw(14) << PSigmanp1[n]["Cus"]->getMedian()
#endif
	 << endl;
    cout << "  Stdev:   "
	 << setw(11) << stdev[n]["Geo"]
	 << setw(14) << stdev[n]["Sca"]
#ifdef CustomModel
	 << setw(14) << stdev[n]["Cus"]
#endif
	 << endl;
    cout << "  SigmaFO: "
	 << setw(11) << osigma[n]
	 << setw(14) << osigma[n]
#ifdef CustomModel
	 << setw(14) << osigma[n]
#endif
	 << endl << endl;
    //
    // save results
    ofile2sum << setw(4)  << n
	      << setw(8)  << sorder[n]
	      << setw(18) << "FixedOrder"
	      << setw(14) << osigma[n]
	      << setw(14) << osigma[n]+CSU_down[n]
	      << setw(14) << osigma[n]+CSU_up[n]
	      << setw(14) << max(fabs(CSU_up[n]),fabs(CSU_down[n]))
	      << endl;
    ofile2sum << setw(4)  << n+0.1
	      << setw(8)  << sorder[n]
	      << setw(18) << "GeometricModel"
	      << setw(14) << PSigmanp1[n]["Geo"]->getMean()
	      << setw(14) << PSigmanp1[n]["Geo"]->getMode()
	      << setw(14) << PSigmanp1[n]["Geo"]->getMedian()
	      << setw(14) <<     stdev[n]["Geo"]
	      << setw(14) <<  xminus68[n]["Geo"]
	      << setw(14) <<   xplus68[n]["Geo"]
	      << setw(14) <<  xminus95[n]["Geo"]
	      << setw(14) <<   xplus95[n]["Geo"]
	      << endl;
    ofile2sum << setw(4)  << n+0.2
	      << setw(8)  << sorder[n]
	      << setw(18) << "ScaleVarModel"
	      << setw(14) << PSigmanp1[n]["Sca"]->getMean()
	      << setw(14) << PSigmanp1[n]["Sca"]->getMode()
	      << setw(14) << PSigmanp1[n]["Sca"]->getMedian()
	      << setw(14) <<     stdev[n]["Sca"]
	      << setw(14) <<  xminus68[n]["Sca"]
	      << setw(14) <<   xplus68[n]["Sca"]
	      << setw(14) <<  xminus95[n]["Sca"]
	      << setw(14) <<   xplus95[n]["Sca"]
	      << endl;
#ifdef CustomModel
    ofile2sum << setw(4)  << n+0.3
	      << setw(8)  << sorder[n]
	      << setw(18) << "CustomModel"
	      << setw(14) << PSigmanp1[n]["Cus"]->getMean()
	      << setw(14) << PSigmanp1[n]["Cus"]->getMode()
	      << setw(14) << PSigmanp1[n]["Cus"]->getMedian()
	      << setw(14) <<     stdev[n]["Cus"]
	      << setw(14) <<  xminus68[n]["Cus"]
	      << setw(14) <<   xplus68[n]["Cus"]
	      << setw(14) <<  xminus95[n]["Cus"]
	      << setw(14) <<   xplus95[n]["Cus"]
	      << endl;
#endif
  }
  ofile2sum.close();
  gettimeofday(&t,NULL);
  cout << "Step 4 done in " << delta_t(tlast,t) << " s" << endl;
  tlast = t;
  //
  //
  for(int k=0; k<Np; k++) {
    ofile2 << setw(12) << xSigma[k];
    for(int n=0; n<nOrd; n++) {
      ofile2 << setw(14) << PSigmanp1[n]["Geo"]->evalDistr(xSigma[k])
	     << setw(14) << PSigmanp1[n]["Sca"]->evalDistr(xSigma[k])
#ifdef CustomModel
	     << setw(14) << PSigmanp1[n]["Cus"]->evalDistr(xSigma[k]) /PSigmanp1[n]["Cus"]->getNorm()  ///////////////////////// FIX THIS
#else
	     << setw(14) << 0
#endif
	;
    }
    ofile2 << endl;
  }
  ofile2.close();
  gettimeofday(&t,NULL);
  cout << "Step 5 done in " << delta_t(tlast,t) << " s" << endl;
  tlast = t;



  gettimeofday(&t,NULL);
  double realt = delta_t(t0, t);
  double cput  = clock() / (float)CLOCKS_PER_SEC;
  cout << "Real time elapsed:  " << realt << " s" << endl;
  cout << "CPU  time elapsed:  " << cput  << " s" << endl;
  
  return 0;
  
}
