#include "StudyDistribution.hh"
#include <vector>
#include <iomanip>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_errno.h>
#include <cuba.h>
#include "integration.hh"

using namespace std;



double FalsiZero(double f(double, void*), double a0, double a1, void *par) {
  double tolerance = 1e-10;
  int maxsteps = 999;
  int side=0;
  double r, fr, fs = f(a0,par), ft = f(a1,par);
  //cout << a0 << "  " << a1 << "  " << fs << "  " << ft << endl;
  if(fs*ft > 0) {
    cout << "ERROR:  Falsi Method works only if the two initial guesses sorround the root." << endl;
    //exit(-35);
    return __builtin_nan("");
  }
  for (int n=0; n<maxsteps; n++) {
    r = (fs*a1 - ft*a0) / (fs - ft);
    if (fabs(a1-a0) < tolerance*fabs(a1+a0)) return r;  // this does not check if f(r)=0, but just that r does not change
    fr = f(r,par);
    //cout << n << "   " << r << "   " << fr << endl;
    if (fr * ft > 0) {
      a1 = r;
      ft = fr;
      if (side==-1) fs /= 2.;
      side = -1;
    }
    else if (fs * fr > 0) {
      a0 = r;
      fs = fr;
      if (side==+1) ft /= 2;
      side = +1;
    }
    else  // fr * f_ very small (looks like zero)
      return r;
  }
  cout << "ERROR:  Falsi Method for finding root did not converge." << endl;
  //exit(-34);
  return __builtin_nan("");
}




//
//	Constructor
//
StudyDistribution::StudyDistribution(double *x, double *distr, unsigned int nPoints) {
  fNpoints = nPoints;
  fx = new double[fNpoints];
  fP = new double[fNpoints];
  for(unsigned int i=0; i<fNpoints; i++) {
    fx[i] = x[i];
    fP[i] = distr[i];
  }
  InitSpline();
}
StudyDistribution::StudyDistribution(vector<double> x, vector<double> distr) {
  fNpoints = x.size();
  fx = new double[fNpoints];
  fP = new double[fNpoints];
  for(unsigned int i=0; i<fNpoints; i++) {
    fx[i] = x[i];
    fP[i] = distr[i];
  }
  InitSpline();
}
StudyDistribution::StudyDistribution(map<double,double> distr) {
  fNpoints = distr.size();
  fx = new double[fNpoints];
  fP = new double[fNpoints];
  int i=0;
  for(map<double,double>::iterator it=distr.begin(); it!=distr.end(); it++) {
    fx[i] = it->first;
    fP[i] = it->second;
    i++;
  }
  InitSpline();
}
void StudyDistribution::InitSpline() {
  //fSpline = gsl_spline_alloc(gsl_interp_cspline, fNpoints);
  fSpline = gsl_spline_alloc(gsl_interp_linear, fNpoints);
  gsl_spline_init(fSpline, fx, fP, fNpoints);
  acc = gsl_interp_accel_alloc ();
  // compute mode
  int imax;
  Max(fP, fNpoints, &imax);
  fMode = fx[imax];
  // compute Norm, Mean and Median
  fNorm = computeIntegral(0, 0);
  fMean = computeIntegral(1, 0)/fNorm;
  fMedian = computeMedian();
  if(fabs(fNorm-1)>0.01) cout << "Warning: in this distribution the normalization " << fNorm << " is significantly different from 1" << endl;
  //
  return;
}
void StudyDistribution::Normalize() {
  for(unsigned int i=0; i<fNpoints; i++) {
    fP[i] /= fNorm;
  }
  gsl_spline_free(fSpline);
  gsl_interp_accel_free (acc);
  InitSpline();
  //fNorm = 1;
  //fMedian = computeMedian();
  return;
}

//
//	Destructor
//
StudyDistribution::~StudyDistribution() {
  delete[] fx;
  delete[] fP;
  gsl_spline_free(fSpline);
  gsl_interp_accel_free (acc);
}

//
//	Copy Constructor
//
//StudyDistribution::StudyDistribution(const StudyDistribution& other):fNOrders(other.fNOrders) {
//  fNOrders = other.fNOrders;
//  fobs = new Observable(*other.fobs);
//  fpriortype = other.fpriortype;
//  fpriormumin = other.fpriormumin;
//  fpriormumax = other.fpriormumax;
//  fNmuPoints = other.fNmuPoints;
//}

StudyDistribution operator + (const StudyDistribution &a, const StudyDistribution &b) {
  unsigned int nPoints = a.fNpoints;
  if(nPoints != b.fNpoints) {
    cout<< "Error: the number of points of the two probability distributions you want to sum is not the same" << endl;
    exit(21);
  }
  double *x = new double[nPoints];
  double *P = new double[nPoints];
  for(unsigned int i=0; i<nPoints; i++) {
    x[i] = a.fx[i];
    if(x[i] != b.fx[i]) {
      cout<< "Error: the sampling points of the two probability distributions you want to sum do not coincide" << endl;
      exit(22);
    }
    P[i] = (a.fP[i] + b.fP[i])/2;
  }
  StudyDistribution distr(x,P,nPoints);
  return distr;
}



double StudyDistribution::evalDistr(double x) {
  if(x<getXmin()) return fP[0];
  if(x>getXmax()) return fP[fNpoints-1];
  return gsl_spline_eval(fSpline, x , acc);
}


double EintegrandGauss(double x, void *par) {
  StudyDistribution *P = (StudyDistribution *) par;
  int mom = P->fMom;
  double x0 = P->fx0;
  double res = P->evalDistr(x) * pow(x-x0,mom);
  return res;
}

double StudyDistribution::computeIntegral(int mom, double x0, double xmin, double xmax) {
  fMom = mom;
  fx0 = x0;
  double res = 0;
  double prec = 1e-3, error;
  //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
  //gsl_set_error_handler_off();
  res = integration::gauss(EintegrandGauss, xmin, xmax, prec, &error, this);
  //gsl_set_error_handler (old_handler);
  return res;
}


double StudyDistribution::Moment(int mom, double x0) {
  return computeIntegral(mom, x0)/fNorm;
}

double StudyDistribution::evalProb(double x1, double x2) {
  double res = 0;
  res = computeIntegral(0, 0, x1, x2);
  return res;
}

double zeroFinding(double x, void *par) {
  StudyDistribution *P = (StudyDistribution *) par;
  double xcen  = P->fxcen;
  double DoB = P->fDoB;
  int direction = P->fDirection;
  double norm  = P->fTrickNorm;
  double res = 0;
  if(direction > 0) res = P->computeIntegral(0, 0, xcen, x)/norm - DoB/2;
  if(direction < 0) res = P->computeIntegral(0, 0, x, xcen)/norm - DoB/2;
  return res;
}

void StudyDistribution::DoBinterval(double xcen, double DoB, double &xplus, double &xminus) {
  fxcen = xcen;
  fDoB = DoB;
  fTrickNorm = 1.; // use the actual area of the curve
  fDirection =  1;
  xplus  = FalsiZero(zeroFinding, fxcen, getXmax(), this);
  if(xplus!=xplus) xplus = getXmax();
  fDirection = -1;
  xminus = FalsiZero(zeroFinding, getXmin(), fxcen, this);
  if(xminus!=xminus) xminus = getXmin();
  return;
}
double StudyDistribution::computeMedian() {
  fxcen = getXmin();
  fDoB = 1;
  fDirection = 1;
  fTrickNorm = fNorm; // renormalize, so that if the distribution is not "complete" but it is symetric, the median will be "more" correct
  return FalsiZero(zeroFinding, fxcen, getXmax(), this);
}



