#include "THunc_model.hh"
#include <vector>
#include <iomanip>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_bessel.h>
#include <cuba.h>
#include "integration.hh"
#include "ObservableScaleDep.hh"

using namespace std;



AuxiliaryObservable::AuxiliaryObservable(QCDobservable *obs, double mu) {
  fobs = obs;
  fmu = mu;
  isInitialized = false;
}
void AuxiliaryObservable::Init() {
  fnpoints = 1200;
  fdmax = 6;
  fnOrders = fobs->GetNumOrders();
  fdeltanp1 = new double*[fnOrders];
  frhatnp1  = new double*[fnOrders];
  double *sigma1 = new double[fnOrders];
  double *sigma2 = new double[fnOrders+1];
  double *sigma3 = new double[fnOrders+1];
  double *r      = new double[fnOrders+1];
  fobs->eval(fmu, sigma1);
  for(int n=0; n<fnOrders; n++) {
    fdeltanp1[n] = new double[fnpoints];
    frhatnp1 [n] = new double[fnpoints];
    sigma2[n] = sigma1[n];
    for(int i=0; i<fnpoints; i++) {
      fdeltanp1[n][i] = -fdmax + i/(fnpoints-1.)*2.*fdmax; // range [-fdmax,fdmax]
      sigma2[n+1] = sigma1[n] + sigma1[0]*fdeltanp1[n][i];      
      QCDobservable *obsloc = new QCDobservable(sigma2, n+2, fobs->GetOffset(), fmu, fobs->alphas(fmu), fobs->GetNf());
      obsloc->eval(fmu, sigma3, r);
      frhatnp1[n][i] = r[n+1]/sigma3[n+1];    
    }
  }
  delete[] sigma1;
  delete[] sigma2;
  delete[] sigma3;
  delete[] r;
  isInitialized = true;
}
AuxiliaryObservable::~AuxiliaryObservable() {
  for(int n=0; n<fnOrders; n++) {
    delete[] fdeltanp1[n];
    delete[] frhatnp1 [n];
  }
  delete[] fdeltanp1;
  delete[] frhatnp1;
}
double AuxiliaryObservable::rhat(int k, double delta) {
  if(!isInitialized) Init();
  if(k<1) exit(45);
  if(delta < fdeltanp1[k-1][0] || delta > fdeltanp1[k-1][fnpoints-1]) {
    cout << "\033[0;31m" << "ERROR: delta = " << delta << " outside of interpolation range [" << fdeltanp1[k-1][0] << ", " << fdeltanp1[k-1][fnpoints-1] << "]" << "\033[0m" << endl;
    exit(24);
  }
  int ii = int((delta+fdmax)/2./fdmax*(fnpoints-1.));
  if(ii==fnpoints-1) ii--;
  //
  return frhatnp1[k-1][ii] + (frhatnp1[k-1][ii+1]-frhatnp1[k-1][ii]) * (delta-fdeltanp1[k-1][ii])/(fdeltanp1[k-1][ii+1]/fdeltanp1[k-1][ii]);
}



///////////
///////////  This algorithm seems to work, but some tests are needed....
///////////
// vector<int> ak_list_old(vector<double> delk, double *ak) {  // the definition of ak is shifted by a unity wrt the paper
//   unsigned int n = delk.size()-1;  // this is n
//   vector<int> goodk;
//   for(int k=0; k<n; k++) {
//     ak[k] = fabs(delk[k+1]/delk[k]);
//     if(ak[k]!=ak[k]) cout << "Potential issue: a_k = " << ak[k] << endl;
//     if(1/ak[k]==0) cout << "Potential issue: a_k = " << ak[k] << endl;
//     // add a check if the ratio is infinite  !!!!!!!!
//     goodk.push_back(k);
//   }
//   ak[n] = 0;  // this is the last one, added by hand. The size of ak is n+1 (from 0 to n)
//   goodk.push_back(n);
//   //
//   vector<int> badk, badj;
//   for(int k=0; k<n; k++) {
//     if(ak[k] <= ak[k+1]) {
//       badk.push_back(k);
//       badj.push_back(1); // this measures how many consecutive bad k's there are
//     }
//   }
//   //
//   for(int i=0; i<badk.size(); i++) {
//     if(i>=badk.size()) break;  // this is needed because in the loop I may remove some elements from the vector
//     int k = badk[i];
//     if(i>0) { // the first check is already done
//       if(k<badk[i-1]+badj[i-1]) { // this badk belongs to the previous bunch
// 	badk.erase(badk.begin()+i);
// 	badj.erase(badj.begin()+i);
// 	i--; // after removing element i, the (i+1)-th (namely the next to check) has become i-th, so I must correct
// 	continue;
//       }
//     }
//     for(int j=2; j<n-k; j++) { // we already know that ak[k+1] is larger than ak[k], so we start from ak[k+2]
//       if(pow(fabs(delk[k+j]/delk[k]),1./j) <= ak[k+j]) { // check if other subsequent ak are larger
// 	badj[i]++;
//       } else {  // otherwise stop checking
// 	break;
//       }
//     }
//     int mj=0;
//     for(int j=-1; j>-k; j--) { // now check backward (the new ak[k] could become larger than the previous ones)
//       if(pow(fabs(delk[k+badj[i]+1]/delk[k+1+j]),1./(1.+badj[i]-1.-j)) >= ak[k+j]) { // check if previous ak are smaller
// 	mj++;
//       } else {  // otherwise stop checking
// 	break;
//       }
//     }
//     badk[i] -= mj;
//     badj[i] += mj;
//   }
//   //
//   for(int i=badk.size()-1; i>=0; i--) {
//     int k = badk[i];
//     int j = badj[i];
//     double aknew = pow(fabs(delk[k+j+1]/delk[k]),1./(1.+j));
//     for(int p=k; p<=k+j; p++) {
//       ak[p] = aknew;
//     }
//     goodk.erase(goodk.begin()+k+1, goodk.begin()+k+j+1);
//   }
//   return goodk;
// }



///////////
///////////  This algorithm is much better, use this
///////////
vector<int> ak_list(vector<double> delk, double *ak) {  // the definition of ak is shifted by a unity wrt the paper
  unsigned int n = delk.size()-1;  // this is n
  delk.push_back(0); // trick to have a fake delta_n+1=0
  vector<int> goodk;
  //bool *kgood = new bool[n+1];
  for(int k=0; k<n; k++) {
    ak[k] = fabs(delk[k+1]/delk[k]); // guesses for the ak
    if(ak[k]!=ak[k]) cout << "Potential issue: k = " << k << "  a_k(guess) = " << ak[k] << "  delta_k = " << delk[k] << "  delta_k+1 = " << delk[k+1] << endl;
    if(1/ak[k]==0) {
      //cout << "Potential issue on a_k for n="<<n<<": a_" << k << " = " << ak[k] << endl;
      ak[k] = std::numeric_limits<double>::max();
    }
    // add a check if the ratio is infinite  !!!!!!!!
    goodk.push_back(k);
    //kgood[k] = true;
  }
  // it's better to set these by hand rather than using the loop above, because if delk[n-1]=0 we would get ak[n]=nan
  ak[n] = 0;  // this is the last one, added by hand. The size of ak is n+1 (from a_0 to a_n)
  goodk.push_back(n);  // at the moment goodk[k] = k
  //kgood[n] = true;
  //
  //if(n==4) {
  //  for(int k=0; k<=n; k++) {
  //    cout << setw(10) << ak[k];
  //  }
  //  cout << endl;
  //}
  //
  int bad = 1;  // it forces at least a run of the while loop
  while(bad>0) {
    bad = 0;
    for(int i=0; i<goodk.size()-1; i++) {
      int k   = goodk[i];
      int kp1 = goodk[i+1]; // this is not k+1, but the next value of k
      if(!(ak[k] > ak[kp1])) { // using this formulation avoids the problems found if some ak are nan
	int kp2;
	if(i==goodk.size()-2) { // we are at the end of the list, there is no goodk[i+2]
	  kp2 = n+1;
	} else {
	  kp2 = goodk[i+2];
	}
	ak[k] = pow(fabs(delk[kp2]/delk[k]),1./(kp2-k));
	//
	for(int j=1; j<kp2-k; j++) ak[k+j] = ak[k];
	goodk.erase(goodk.begin()+i+1);
	//kgood[kp1] = false;
	bad++; // this tells the cycle that something has changed in the list, so it must be checked again
	break; // this may slow down the checks if the list is long (never...), but it ensures that no problems are generated
      }
    }
  }
  //if(n==4) {
  //  for(int k=0; k<=n; k++) {
  //    cout << setw(10) << ak[k];
  //  }
  //  cout << endl;
  //  int kk=0;
  //  for(int k=0; k<=n; k++) {
  //    cout << setw(10) << (kgood[k] ? goodk[kk] : -1);
  //    if(kgood[k]) kk++;
  //  }
  //  cout << endl << endl;
  //}
  //
  return goodk;
}










//////////////
// Constructor
//////////////
THuncModel::THuncModel(double *sigma, int nOrders, double probHpm) {
  fNOrders = nOrders;
  fdSigma  = new double [fNOrders];
  fdelta   = new double [fNOrders];
  fdSigma[0] = sigma[0];
  fdelta [0] = 1;
  for (int i=1; i<fNOrders; i++) {
    fdSigma[i] = sigma[i] - sigma[i-1];
    fdelta [i] = fdSigma[i]/sigma[0];
  }
  SetPHplusPHminus(probHpm,probHpm);
  fdelint = NULL;
  frhint  = NULL;
  EIS = false;
}
THuncModel::THuncModel(double *sigma, double *r, int nOrders, double probHpm) {
  fNOrders = nOrders;
  fdSigma  = new double [fNOrders];
  fdelta   = new double [fNOrders];
  frk      = new double [fNOrders];
  frhatk   = new double [fNOrders];
  for (int i=0; i<fNOrders; i++) {
    if(i==0) {
      fdSigma[0] = sigma[0];
      fdelta [0] = 1;
    } else {
      fdSigma[i] = sigma[i] - sigma[i-1];
      fdelta [i] = fdSigma[i]/sigma[0];
    }
    frk    [i] = r[i];
    frhatk [i] = r[i]/fabs(sigma[i]);
  }
  if(frk[0]==0) {
    frk    [0] = sigma[0]/2;
    frhatk [0] = 0.5;
  }
  SetPHplusPHminus(probHpm,probHpm);
  fdelint = NULL;
  frhint  = NULL;
  EIS = false;
}
THuncModel::THuncModel(QCDobservable *obs, double mu, double probHpm) {
  auxObs = new AuxiliaryObservable(obs, mu);
  fNOrders = obs->GetNumOrders();
  fdSigma  = new double [fNOrders];
  fdelta   = new double [fNOrders];
  frk      = new double [fNOrders];
  frhatk   = new double [fNOrders];
  double *sigma = new double[fNOrders];
  obs->eval(mu, sigma, frk);
  if(frk[0]==0) frk[0] = sigma[0]/2;
  for (int i=0; i<fNOrders; i++) {
    if(i==0) {
      fdSigma[0] = sigma[0];
      fdelta [0] = 1;
    } else {
      fdSigma[i] = sigma[i] - sigma[i-1];
      fdelta [i] = fdSigma[i]/sigma[0];
    }
    frhatk [i] = frk[i]/fabs(sigma[i]);
  }
  SetPHplusPHminus(probHpm,probHpm);
  fdelint = NULL;
  frhint  = NULL;
  EIS = false;
}
//////////////
// Destructor
//////////////
THuncModel::~THuncModel() {
  delete [] fdSigma;
  delete [] fdelta;
}
//////////////
// Copy Constructor
//////////////
THuncModel::THuncModel(const THuncModel& other) {
  fNOrders = other.fNOrders;
  fdSigma = new double [fNOrders];
  fdelta  = new double [fNOrders];
  frk      = new double [fNOrders];
  frhatk   = new double [fNOrders];
  for (int i=1; i<fNOrders; i++) {
    fdSigma[i] = other.fdSigma[i];
    fdelta [i] = other.fdelta [i];
    frk    [i] = other.frk    [i];
    frhatk [i] = other.frhatk [i];
  }
  fPHplus  = other.fPHplus;
  fPHminus = other.fPHminus;
  EIS = other.EIS;
  fOffset = other.fOffset;
  fmu = other.fmu;
  fas = other.fas;
  fnf = other.fnf;
}

void THuncModel::checkN(int n) {
  if(n+1>fNOrders) {
    cout << "Error: you are using a too large value of n (n=" << n << " > nOrders-1=" << fNOrders-1 << ")" << endl;
    exit(7);
  }
  return;
}

void THuncModel::SetPHplusPHminus(double probHplus, double probHminus) {
  if(probHplus>1||probHplus<0) {
    cout << "Setting the probability of the H+ hypotesis outside the [0,1] range." << endl;
    exit(21);
  }
  if(probHminus>1||probHminus<0) {
    cout << "Setting the probability of the H- hypotesis outside the [0,1] range." << endl;
    exit(22);
  }
  if(probHplus+probHminus>1) {
    cout << "Setting a negative probability for the H0 hypotesis." << endl;
    exit(23);
  }
  fPHplus  = probHplus;
  fPHminus = probHminus;
  return;
}

double THuncModel::P_deltanp1(double delta_np1, int n) {
  checkN(n);
  return P_deltanp1_dots_delta1(delta_np1,n)/P_deltan_dots_delta1(n);
}
double THuncModel::P_Sigmanp1(double Sigma_np1, int n) {
  checkN(n);
  double delta_np1 = Sigma_np1/fdSigma[0];
  for(int i=0; i<=n; i++) delta_np1 -= fdelta[i];
  return P_deltanp1(delta_np1,n)/fdSigma[0];
}

double integrandGauss(double delta_np1, void *par) {
  THuncModel *m = (THuncModel *) par;
  double delta_np2 = m->fDnp - delta_np1;
  int n = m->fNlocal;
  return m->P_deltanp2_dots_delta1(delta_np2,delta_np1,n);
}
double THuncModel::P_Sigmanp2(double Sigma_np2, int n) {
  checkN(n);
  fNlocal = n;
  fDnp = Sigma_np2/fdSigma[0];
  for(int i=0; i<=n; i++) fDnp -= fdelta[i];
  double res = 0;
  double prec = 3e-3, error;
  gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
  gsl_set_error_handler_off();
  res = integration::gauss(integrandGauss, -6, 6, prec, &error, this);
  gsl_set_error_handler (old_handler);
  return res/P_deltan_dots_delta1(n)/fdSigma[0];
}
int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
  THuncModel *m = (THuncModel *) pars;
  double range_width = 12;
  double delta_np1 = range_width*(x[0]-0.5);
  double delta_np2 = range_width*(x[1]-0.5);
  double delta_np3 = m->fDnp - delta_np2 - delta_np1;
  int n = m->fNlocal;
  res[0] = m->P_deltanp3_dots_delta1(delta_np3,delta_np2,delta_np1,n);
  res[0] *= range_width*range_width;
  return 0;
}
double THuncModel::P_Sigmanp3(double Sigma_np3, int n) {
  checkN(n);
  fNlocal = n;
  fDnp = Sigma_np3/fdSigma[0];
  for(int i=0; i<=n; i++) fDnp -= fdelta[i];
  double cres[1], error[1], prob[1];
  double epsrel=1e-2, epsabs=1.e-6;
  int mineval = 0, maxeval = 50000;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail, seed;
  nregions = 0;
  seed = 0;
  //system("export CUBACORES=8");
  int dim = 2, ncomp = 1;
  Cuhre(dim, ncomp, integrandCUBA, this, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
  //Vegas(dim, ncomp, integrandCUBA, this, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, 50000, 20000, 20000, 1, NULL, NULL, &neval, &fail, cres, error, prob);
  //  
  if(fail == -1) {
    cout << "\033[0;31m  ERROR: dimension out of range.\033[0m" << endl;
  } else if(fail>0) {
    cout << "\033[0;31m  ERROR: precision not reached.\033[0m" << endl;
  }
  return cres[0]/P_deltan_dots_delta1(n)/fdSigma[0];
}


// this treatment of the sign is a hack such that P(deltanpk...deltanp1 | deltan...delta1) is correct acting only on the numerator P(deltanpk...delta1) but not on the denominator P(deltan...delta1)
double THuncModel::P_deltan_dots_delta1(int n) {
  return P_deltan_dots_delta1_H0(n);
}
double THuncModel::P_deltanp1_dots_delta1(double delta_np1, int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  double res = 1;
  int Hoffset = 0;
  bool Hplus  = AllSignsPlus(vec);
  bool Hminus = AlternatingSigns(vec, Hoffset);
  if(Hplus && fPHplus>0) {
    double BF = (1-fPHplus)/fPHplus/pow(2,n);  // Bayes factor P0(H0)/P0(H+) divided by 2^n
    res *= (2*theta(delta_np1)+BF)/(1+BF);
  }
  else if(Hminus && fPHminus>0) {
    double BF = (1-fPHminus)/fPHminus/pow(2,n);  // Bayes factor P0(H0)/P0(H+) divided by 2^n
    res *= (2*theta(pow(-1.,n+1+Hoffset)*delta_np1)+BF)/(1+BF);
  }
  res *= P_deltanp1_dots_delta1_H0(delta_np1,n);
  return res;
}
double THuncModel::P_deltanp2_dots_delta1(double delta_np2, double delta_np1, int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  double res = 1;
  int Hoffset = 0;
  bool Hplus  = AllSignsPlus(vec);
  bool Hminus = AlternatingSigns(vec, Hoffset);
  if(Hplus && fPHplus>0) {
    double BF = (1-fPHplus)/fPHplus/pow(2,n);  // Bayes factor P0(H0)/P0(H+) divided by 2^n
    res *= (4*theta(delta_np2)*theta(delta_np1)+BF)/(1+BF);
  }
  else if(Hminus && fPHminus>0) {
    double BF = (1-fPHminus)/fPHminus/pow(2,n);  // Bayes factor P0(H0)/P0(H+) divided by 2^n
    res *= (4*theta(pow(-1.,n+2+Hoffset)*delta_np2)*theta(pow(-1.,n+1+Hoffset)*delta_np1)+BF)/(1+BF);
  }
  res *= P_deltanp2_dots_delta1_H0(delta_np2,delta_np1,n);
  return res;
}
double THuncModel::P_deltanp3_dots_delta1(double delta_np3, double delta_np2, double delta_np1, int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  double res = 1;
  int Hoffset = 0;
  bool Hplus  = AllSignsPlus(vec);
  bool Hminus = AlternatingSigns(vec, Hoffset);
  if(Hplus && fPHplus>0) {
    double BF = (1-fPHplus)/fPHplus/pow(2,n);  // Bayes factor P0(H0)/P0(H+) divided by 2^n
    res *= (8*theta(delta_np3)*theta(delta_np2)*theta(delta_np1)+BF)/(1+BF);
  }
  else if(Hminus && fPHminus>0) {
    double BF = (1-fPHminus)/fPHminus/pow(2,n);  // Bayes factor P0(H0)/P0(H+) divided by 2^n
    res *= (8*theta(pow(-1.,n+3+Hoffset)*delta_np3)*theta(pow(-1.,n+2+Hoffset)*delta_np2)*theta(pow(-1.,n+1+Hoffset)*delta_np1)+BF)/(1+BF);
  }
  res *= P_deltanp3_dots_delta1_H0(delta_np3,delta_np2,delta_np1,n);
  return res;
}
void THuncModel::computeScaleDependence(vector<double> delta, double *rhat) {
  int m = delta.size()-1;
  double *sigma = new double[m+1];
  double *osigm = new double[m+1];
  //rhat          = new double[m+1];
  sigma[0] = fdSigma[0];
  for(int k=1; k<=m; k++) sigma[k] = sigma[k-1] + sigma[0]*delta[k];
  QCDobservable obs(sigma, m+1, fOffset, fmu, fas, fnf);
  obs.eval(fmu,osigm,rhat);
  for(int k=0; k<=m; k++) rhat[k] /= fabs(sigma[k]);
  if(rhat[0]==0) rhat[0] = 0.5;
  //for(int k=0; k<=n; k++) cout << k << "  " << frhatk[k]/rhatk[k] << endl;
  //cout << n << setw(14) << fdelta[n] << setw(14) << rhatk[n]
  //     << setw(24) << delta_np1 << setw(14) << rhatk[n+1] << setw(16) << fabs(delta_np1)/rhatk[n]
  //     << setw(24) << delta_np2 << setw(16) << fabs(delta_np2)/rhatk[n+1]
  //     << endl;
  return;
}










///////////////////////////////
//  Geometric behaviour Model
///////////////////////////////

double aintegral(double akm1, double ak, int k, unsigned int n, double eps, int cp, int ap) {
  double den = 1-n*(1.+n)/2.+ap+k*(n+eps-cp);
  if(den==0) return log(Min(1.,akm1)/Min(1.,ak));
  return ( pow(Min(1.,akm1),den) - pow(Min(1.,ak),den) ) /den;
}
double P_deltam_dots_delta1(vector<double> delk, double eps, int cp=0, int ap=0) {  // cp and ap are the extra contributions to the powers of c and a needed for the variance
  unsigned int n = delk.size()-1;  // the largest is delta_n, so there are n+1 delta's
  double *ak = new double[n+1];  // I need as well n+1 a_k's, from a_0 to a_n
  vector<int> goodk;
  goodk = ak_list(delk, ak);
  if(goodk[0]!=0) cout << "Can this be correct?" << endl;
  //double res = aintegral(1,ak[0],0,n,eps,cp,ap);  // here ak[0] should be ak[goodk[0]], but it must be the same
  double res = 2*(aintegral(1,ak[0],0,n,eps,cp,ap)-aintegral(1,ak[0],0,n,eps,cp,ap+1));
  for(int i=1; i<goodk.size(); i++) {
    int k   = goodk[i];
    int km1 = goodk[i-1];
    //cout << k << "  " << ak[k]<< endl;
    //res += aintegral(ak[km1],ak[k],k,n,eps,cp,ap) / pow(fabs(delk[k]),n+eps-cp);
    res += 2*(aintegral(ak[km1],ak[k],k,n,eps,cp,ap)-aintegral(ak[km1],ak[k],k,n,eps,cp,ap+1)) / pow(fabs(delk[k]),n+eps-cp);
    //cout << res << endl;
  }
  //if(delk[delk.size()-1]==0) {
    //for(int i=0; i<goodk.size(); i++) cout << ak[goodk[i]] << "  ";
    //cout << res << endl;
  //}
  return res*eps/(n+eps-cp)/pow(2.,n);
}

double THuncModel_Geo::P_deltan_dots_delta1_H0(int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  return ::P_deltam_dots_delta1(vec,feps);
}

double THuncModel_Geo::P_deltanp1_dots_delta1_H0(double delta_np1, int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  vec.push_back(delta_np1);
  return ::P_deltam_dots_delta1(vec,feps);
}

double THuncModel_Geo::P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  vec.push_back(delta_np1);
  vec.push_back(delta_np2);
  return ::P_deltam_dots_delta1(vec,feps);
}

double THuncModel_Geo::P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  vec.push_back(delta_np1);
  vec.push_back(delta_np2);
  vec.push_back(delta_np3);
  return ::P_deltam_dots_delta1(vec,feps);
}

double THuncModel_Geo::Var_deltanp1(int n) {
  checkN(n);
  vector<double> vec(fdelta, fdelta + n+1);
  double res = ::P_deltam_dots_delta1(vec,feps,2,2*(n+1))/::P_deltam_dots_delta1(vec,feps) /3.;
  return res;
}

double THuncModel_Geo::Prior(double c, double a) {
  if(a<0) return 0;
  if(a>1) return 0;
  if(c<1) return 0;
  //return feps/pow(c,1+feps);
  return feps/pow(c,1+feps) * 2*(1-a);
}

double THuncModel_Geo::P_c_a_H0(double c, double a, int n) {
  double *maxvec = new double[n];
  for(int i=1; i<=n; i++) {
    maxvec[i-1] = fabs(fdelta[i])/pow(a,i);
  }
  vector<double> vec(fdelta, fdelta + n+1);
  return Prior(c,a) * theta(c-Max(maxvec,n)) / pow(c,n) / pow(a,n*(n+1)/2.) / pow(2,n) / ::P_deltam_dots_delta1(vec,feps);
}


// void THuncModel_Geo::test() {
//   vector<double> vec(fdelta, fdelta + fNOrders);
//   //vec.push_back(0.0384754);
//   double *vec2 = new double[vec.size()];
//   vector<int> goodk;
//   goodk = ak_list(vec, vec2);
//   for(int k=0; k<vec.size(); k++) {
//     cout << k << "  " << vec2[k] << endl;
//   }
//   for(int i=0; i<goodk.size(); i++) {
//     int k = goodk[i];
//     cout << k << "  " << vec2[k] << endl;
//   }
//   double res = ::P_deltam_dots_delta1(vec,feps);
//   cout << "res = " << res << endl;
// }












///////////////////////////////
//  Scale variation Model
///////////////////////////////

double ScaP_deltam_dots_delta1(vector<double> vec, double *rk, double gamma) {
  double lamn = Max(vec);
  int m = vec.size();
  double res = 1;
  for(int k=0; k<m; k++) res /= 2*rk[k];
  gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
  gsl_set_error_handler_off();
  if(m>0) res *= gsl_sf_gamma_inc(gamma-m+1,lamn);
  gsl_set_error_handler (old_handler);
  return res/gsl_sf_gamma(1+gamma);
}
double THuncModel_Scale::P_deltan_dots_delta1_H0(int n) {
  checkN(n);
  vector<double> vec;
  for(int k=0; k<n; k++) {
    vec.push_back(fabs(fdelta[k+1])/frhatk[k]);
  }
  return ScaP_deltam_dots_delta1(vec, frhatk, fgamma);
}
double THuncModel_Scale::P_deltanp1_dots_delta1_H0(double delta_np1, int n) {
  checkN(n);
  vector<double> vec;
  for(int k=0; k<n; k++) {
    vec.push_back(fabs(fdelta[k+1])/frhatk[k]);
  }
  vec.push_back(fabs(delta_np1)/frhatk[n]);
  return ScaP_deltam_dots_delta1(vec, frhatk, fgamma);
}

double THuncModel_Scale::P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n) {
  checkN(n);
  ///////////////////  compute the next rk
  double *rhatk = new double[n+2];
  for(int k=0; k<=n; k++) rhatk[k] = frhatk[k];
  rhatk[n+1] = auxObs->rhat(n+1,delta_np1);
  ///////////////////
  vector<double> vec;
  for(int k=0; k<n; k++) {
    vec.push_back(fabs(fdelta[k+1])/rhatk[k]);
  }
  vec.push_back(fabs(delta_np1)/rhatk[n]);
  vec.push_back(fabs(delta_np2)/rhatk[n+1]);
  return ScaP_deltam_dots_delta1(vec, rhatk, fgamma);
}
double THuncModel_Scale::P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n) {
  checkN(n);
  ///////////////////  compute the next rk
  vector<double> delta;
  for(int k=0; k<=n; k++) delta.push_back(fdelta[k]);
  delta.push_back(delta_np1);
  delta.push_back(delta_np2);
  double *rhatk = new double[delta.size()];
  computeScaleDependence(delta, rhatk);
  ///////////////////
  vector<double> vec;
  for(int k=0; k<n; k++) {
    vec.push_back(fabs(fdelta[k+1])/rhatk[k]);
  }
  vec.push_back(fabs(delta_np1)/rhatk[n]);
  vec.push_back(fabs(delta_np2)/rhatk[n+1]);
  vec.push_back(fabs(delta_np3)/rhatk[n+2]);
  return ScaP_deltam_dots_delta1(vec, rhatk, fgamma);
}

double THuncModel_Scale::Prior(double lambda) {
  if(lambda<0) return 0;
  return pow(lambda,fgamma)*exp(-lambda) /gsl_sf_gamma(1+fgamma);
}
double THuncModel_Scale::P_lambda_H0(double lambda, int n) {
  if(n==0) return Prior(lambda);
  checkN(n);
  vector<double> vec;
  for(int k=0; k<n; k++) vec.push_back(fabs(fdelta[k+1])/frhatk[k]);
  double lamn = Max(vec);
  return Prior(lambda) * theta(lambda-lamn) / pow(lambda,n) / gsl_sf_gamma_inc(fgamma-n+1,lamn) * gsl_sf_gamma(1+fgamma);
}

















///////////////////////////////
//  Customizeable Model
///////////////////////////////

double THuncModel_Custom::P_deltan_dots_delta1_H0(int n) {
  checkN(n);
  fNlocal = n;
  if(fdelint!=NULL) delete[] fdelint;
  if(frhint !=NULL) delete[] frhint;
  fdelint = new double[fNlocal+1];
  frhint  = new double[fNlocal+1];
  for(int k=0; k<=n; k++) fdelint[k] = fdelta[k];
  //vector<double> delta(fdelint, fdelint + n+1);
  //computeScaleDependence(delta, frhint);
  for(int k=0; k<=n; k++) frhint[k] = frhatk[k];
  double res = computeIntegral();
  return res;
}
double THuncModel_Custom::P_deltanp1_dots_delta1_H0(double delta_np1, int n) {
  checkN(n);
  fNlocal = n+1;
  if(fdelint!=NULL) delete[] fdelint;
  if(frhint !=NULL) delete[] frhint;
  fdelint = new double[fNlocal+1];
  frhint  = new double[fNlocal+1];
  for(int k=0; k<=n; k++) fdelint[k] = fdelta[k];
  fdelint[n+1] = delta_np1;
  //vector<double> delta(fdelint, fdelint + n+2);
  //computeScaleDependence(delta, frhint);
  for(int k=0; k<=n; k++) frhint[k] = frhatk[k];
  frhint[n+1] = auxObs->rhat(n+1,delta_np1);
  double res = computeIntegral();
  return res;
}
double THuncModel_Custom::P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n) {
  checkN(n);
  fNlocal = n+2;
  if(fdelint!=NULL) delete[] fdelint;
  if(frhint !=NULL) delete[] frhint;
  fdelint = new double[fNlocal+1];
  frhint  = new double[fNlocal+1];
  for(int k=0; k<=n; k++) fdelint[k] = fdelta[k];
  fdelint[n+1] = delta_np1;
  fdelint[n+2] = delta_np2;
  vector<double> delta(fdelint, fdelint + n+3);
  computeScaleDependence(delta, frhint);
  //for(int k=0; k<=n; k++) frhint[k] = frhatk[k];
  //frhint[n+1] = auxObs->rhat(n+1,delta_np1);
  //frhint[n+2] = auxObs->rhat(n+2,delta_np1,delta_np2);
  double res = computeIntegral();
  return res;
}
double THuncModel_Custom::P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n) {
  checkN(n);
  fNlocal = n+3;
  if(fdelint!=NULL) delete[] fdelint;
  fdelint = new double[fNlocal+1];
  for(int k=0; k<=n; k++) fdelint[k] = fdelta[k];
  fdelint[n+1] = delta_np1;
  fdelint[n+2] = delta_np2;
  fdelint[n+3] = delta_np3;
  vector<double> delta(fdelint, fdelint + n+4);
  if(frhint != NULL) delete[] frhint;
  computeScaleDependence(delta, frhint);
  double res = computeIntegral();
  return res;
}

double THuncModel_Custom::P_pars_H0(int n, const double *par) {
  double res = custom_P0(par);
  if(n==0) return res;
  checkN(n);
  fNlocal = n;
  if(fdelint!=NULL) delete[] fdelint;
  fdelint = new double[fNlocal+1];
  for(int k=0; k<=n; k++) fdelint[k] = fdelta[k];
  res /= computeIntegral();
  for(int k=1; k<=n; k++) res *= custom_P_deltan_given_pars(fdelta, frhatk, k, par);// /norm_custom_P_deltan_given_pars(fdelta, k, par);
  //res /= norm_custom(fdelta, frhatk, n, par);
  return res;
}
double THuncModel_Custom::P_deltan_dots_delta1_custom_integrand(int n, double *delk, double *rhat, const double *par) {
  double res = custom_P0(par);
  for(int k=1; k<=n; k++) res *= custom_P_deltan_given_pars(delk, rhat, k, par);// / norm_custom_P_deltan_given_pars(delk, k, par);
  //res /= norm_custom(delk, rhat, n, par);
  return res;
}
//
int integrandCUBAcustom(const int *ndim, const double par[], const int *ncomp, double res[], void *model) {
  THuncModel_Custom *m = (THuncModel_Custom *) model;
  res[0]  = m->P_deltan_dots_delta1_custom_integrand(m->fNlocal, m->fdelint, m->frhint, par);
  return 0;
}
double THuncModel_Custom::computeIntegral() {
  double cres[1], error[1], prob[1];
  double epsrel=1e-2, epsabs=1.e-8;
  int mineval = 1000, maxeval = 500000;
  int nstart = 20000, nincrease = 40000, nbatch = 1000, gridno = 5;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail, seed;
  nregions = 0;
  seed = 0;
  //system("export CUBACORES=8");
  int dim = fnPar, ncomp = 1;
  if(dim==1) dim=2;
  if(dim<3) Cuhre(dim, ncomp, integrandCUBAcustom, this, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
  else      Vegas(dim, ncomp, integrandCUBAcustom, this, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
  //  
  if(fail == -1) {
    cout << "\033[0;31m  ERROR: dimension out of range.\033[0m" << endl;
  } else if(fail>0) {
    //cout << "\033[0;31m  ERROR: precision not reached.\033[0m" << endl;
  }
  return cres[0];
}
double integrandGaussCustom(double delta_n, void *par) {
  //return 1./12.;
  THuncModel_Custom *m = (THuncModel_Custom *) par;
  int n = m->fnNorm;
  double *del   = new double[n+1];
  double *rhatk = new double[n+1];
  for(int k=0; k<n; k++) del[k] = m->fdeltaNorm[k];
  del[n] = delta_n;
  //vector<double> delta(del, del + n+1);
  //m->computeScaleDependence(delta, rhatk);
  for(int k=0; k<n; k++) rhatk[k] = m->frhatk[k];
  rhatk[n] = m->auxObs->rhat(n,delta_n);
  return custom_P_deltan_given_pars(del, rhatk, n, m->fParsNorm);
}
double THuncModel_Custom::norm_custom_P_deltan_given_pars(double *delta, int n, const double *pars) {
  fnNorm = n;
  if(fdeltaNorm!=NULL) delete[] fdeltaNorm;
  fdeltaNorm = new double[n];
  for(int k=0; k<n; k++) fdeltaNorm[k] = delta[k];
  if(fParsNorm!=NULL) delete[] fParsNorm;
  fParsNorm = new double[fnPar];
  for(int i=0; i<fnPar; i++) fParsNorm[i] = pars[i];
  double res = 0;
  double prec = 3e-3, error;
  //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
  //gsl_set_error_handler_off();
  res = integration::gauss(integrandGaussCustom, -6, 6, prec, &error, this);
  //gsl_set_error_handler (old_handler);
  return res;
}
int integrandCUBAcustomNorm(const int *ndim, const double x[], const int *ncomp, double res[], void *model) {
  THuncModel_Custom *m = (THuncModel_Custom *) model;
  int n = m->fnNorm;
  res[0] = 1;
  for(int k=1; k<=n; k++) {
    double deltak = -6 + x[k-1]*12.;
    double jac = 12;
    double *del  = new double[k+1];
    double *rhat = new double[k+1];
    for(int j=0; j<k; j++) {
      del [j] = m->fdeltaNorm[j];
      rhat[j] = m->frhatkNorm[j];
    }
    del [k] = deltak;
    rhat[k] = m->auxObs->rhat(k,deltak); // I am assuming that the passed delta are the known one, which is no longer the case for Sigma_np2 and beyond
    //    
    res[0] *= custom_P_deltan_given_pars(del, rhat, k, m->fParsNorm) * jac;
  }
  return 0;
}
double THuncModel_Custom::norm_custom(double *delta, double *rhatk, int n, const double *pars) {
  fnNorm = n;
  if(fdeltaNorm!=NULL) delete[] fdeltaNorm;
  if(frhatkNorm!=NULL) delete[] frhatkNorm;
  fdeltaNorm = new double[n];
  frhatkNorm = new double[n];
  for(int k=0; k<n; k++) {
    fdeltaNorm[k] = delta[k];
    frhatkNorm[k] = rhatk[k];
  }
  if(fParsNorm!=NULL) delete[] fParsNorm;
  fParsNorm = new double[fnPar];
  for(int i=0; i<fnPar; i++) fParsNorm[i] = pars[i];
  //
  double cres[1], error[1], prob[1];
  double epsrel=1e-2, epsabs=1.e-8;
  int mineval = 1000, maxeval = 500000;
  int nstart = 20000, nincrease = 40000, nbatch = 1000, gridno = 6;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail, seed;
  nregions = 0;
  seed = 0;
  //system("export CUBACORES=8");
  int dim = fnPar, ncomp = 1;
  if(dim==1) dim=2;
  if(dim<3) Cuhre(dim, ncomp, integrandCUBAcustomNorm, this, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
  else      Vegas(dim, ncomp, integrandCUBAcustomNorm, this, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
  //  
  if(fail == -1) {
    cout << "\033[0;31m  ERROR: dimension out of range.\033[0m" << endl;
  } else if(fail>0) {
    //cout << "\033[0;31m  ERROR: precision not reached.\033[0m" << endl;
  }
  return cres[0];
}



double THuncModel_Custom::P_par_marginal_H0(int n, double par, int whichpar) {
  checkN(n);
  fNlocal = n;
  if(fdelint!=NULL) delete[] fdelint;
  fdelint = new double[fNlocal+1];
  for(int k=0; k<=n; k++) fdelint[k] = fdelta[k];
  fParFixed = par;
  fwhichPar = whichpar;
  double res = computeIntegralMarginal();
  if(n>0) res /= computeIntegral();
  //res /= norm_custom(fdelta, frhatk, n, par);
  return res;
}
double THuncModel_Custom::P_par_marginal_integrand(int n, double *delk, double *rhat, double parFixed, int whichPar, const double *intpar) {
  double *par = new double[fnPar];
  int j=0;
  for(int i=0; i<fnPar; i++) {
    if(i==whichPar) par[i] = parFixed;
    else par[i] = intpar[j++];
  }
  double res = custom_P0(par);
  if(n==0) return res;
  for(int k=1; k<=n; k++) res *= custom_P_deltan_given_pars(delk, rhat, k, par);// /norm_custom_P_deltan_given_pars(fdelta, k, par);
  return res;
}
int integrandCUBAcustomMarginal(const int *ndim, const double par[], const int *ncomp, double res[], void *model) {
  THuncModel_Custom *m = (THuncModel_Custom *) model;
  res[0]  = m->P_par_marginal_integrand(m->fNlocal, m->fdelint, m->frhint, m->fParFixed, m->fwhichPar, par);
  return 0;
}
double THuncModel_Custom::computeIntegralMarginal() {
  double cres[1], error[1], prob[1];
  double epsrel=1e-2, epsabs=1.e-8;
  int mineval = 1000, maxeval = 500000;
  int nstart = 20000, nincrease = 40000, nbatch = 1000, gridno = 5;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail, seed;
  nregions = 0;
  seed = 0;
  //system("export CUBACORES=8");
  int dim = fnPar-1, ncomp = 1;
  if(dim==1) dim=2;
  if(dim<3) Cuhre(dim, ncomp, integrandCUBAcustomMarginal, this, 1, epsrel, epsabs, verbose | last,       mineval, maxeval, 13, NULL, NULL, &nregions, &neval, &fail, cres, error, prob);
  else      Vegas(dim, ncomp, integrandCUBAcustomMarginal, this, 1, epsrel, epsabs, verbose | last, seed, mineval, maxeval, nstart, nincrease, nbatch, gridno, NULL, NULL, &neval, &fail, cres, error, prob);
  //  
  if(fail == -1) {
    cout << "\033[0;31m  ERROR: dimension out of range.\033[0m" << endl;
  } else if(fail>0) {
    //cout << "\033[0;31m  ERROR: precision not reached.\033[0m" << endl;
  }
  return cres[0];
}



