#pragma once

#include <cmath>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include "common.hh"
#include "ObservableScaleDep.hh"

using namespace std;


class AuxiliaryObservable {
public:
  AuxiliaryObservable(QCDobservable *obs, double mu);
  ~AuxiliaryObservable();
  //
  void Init();
  double rhat(int k, double delta);
  //
private:
  bool isInitialized;
  QCDobservable *fobs;
  double fdmax, fmu;
  int fnOrders, fnpoints;
  double **fdeltanp1, **frhatnp1;
};



// mother class

class THuncModel {
public:
  THuncModel(double *sigma,            int nOrder, double probHplusminus=0);
  THuncModel(double *sigma, double *r, int nOrder, double probHplusminus=0);
  THuncModel(QCDobservable *obs, double mu,        double probHplusminus=0);
  THuncModel(const THuncModel& other);
  ~THuncModel();
  //
  int  GetMaxN() { return fNOrders-1; }
  void SetPHplusPHminus(double probHplus=0.333, double probHminus=0.333);
  void SetExtraInfo(double offset, double muR, double asR, int nf) { fOffset=offset; fmu=muR; fas=asR; fnf=nf; EIS=true; }
  //
  // distributions
  double P_Sigmanp3(double Sigma_np3, int n); // this is the probability distribution for Sigma_{n+3}
  double P_Sigmanp2(double Sigma_np2, int n); // this is the probability distribution for Sigma_{n+2}
  double P_Sigmanp1(double Sigma_np1, int n); // this is the probability distribution for Sigma_{n+1}
  //
  double P_deltanp1(double delta_np1, int n); // this is the probability distribution for delta_{n+1}
  double P_deltanp1(double delta_np1) { return P_deltanp1(delta_np1,GetMaxN()); } // the same for maximum n
  //
  // virtual members
  virtual double Var_deltanp1(int n) { return 0; }  // this is the variance of P(delta_{n+1})
  // needed for removing scale uncertainty
  double         P_deltanp3_dots_delta1   (double delta_np3, double delta_np2, double delta_np1, int n);
  virtual double P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n) =0;
  double         P_deltanp2_dots_delta1   (double delta_np2, double delta_np1, int n);
  virtual double P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n) =0;
  double         P_deltanp1_dots_delta1   (double delta_np1, int n);
  virtual double P_deltanp1_dots_delta1_H0(double delta_np1, int n) =0;
  double         P_deltan_dots_delta1   (int n);
  virtual double P_deltan_dots_delta1_H0(int n) =0;

protected:
  int fNOrders;
  double *fdelta, *fdSigma, *frk, *frhatk;
  double fPHplus, fPHminus;
  AuxiliaryObservable *auxObs;
  THuncModel& operator=( const THuncModel& other );
  void checkN(int n);
  friend int integrandCUBAmixed(const int *ndim, const double x[], const int *ncomp, double res[], void *pars);
  friend double integrandGauss(double x, void *par);
  friend int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars);
  void computeScaleDependence(vector<double> delta, double *rhat);
  int fNlocal;
  double *fdelint, *frhint, flam0int, fDnp;
  double fOffset, fmu, fas;
  int fnf;
  bool EIS;
};



// geometric behaviour model
//
class THuncModel_Geo : public THuncModel {
public:
  THuncModel_Geo(double *sigma,            int nOrder, double probHplusminus=0) : THuncModel(sigma,  nOrder,probHplusminus), feps(0.1) {}
  //THuncModel_Geo(double *sigma, double *r, int nOrder, double probHplus=0) : THuncModel(sigma,r,nOrder,probHplus), feps(1) {}
  THuncModel_Geo(QCDobservable *obs, double mu,        double probHplusminus=0) : THuncModel(obs,mu,        probHplusminus), feps(0.1) {}
  THuncModel_Geo(const THuncModel_Geo& other) : THuncModel(other) { feps = other.feps; }
  //
  //double P_deltanp1(double delta_np1, int n); // this is the probability distribution for delta_{n+1}
  double Var_deltanp1(int n);  // this is the variance of the same distribution
  double P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n);
  double P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n);
  double P_deltanp1_dots_delta1_H0(double delta_np1, int n);
  double P_deltan_dots_delta1_H0(int n);
  //
  void SetPrior(double eps=1) { feps = eps; }
  //
  double P_c_a_H0(double c, double a, int n);
  //void test();
protected:
  double feps;
  double Prior(double c, double a);
};


// scale variation model
//
class THuncModel_Scale : public THuncModel {
private:
  double computeIntegral();
public:
  //THuncModel_Scale(double *sigma,            int nOrder, double probHplus=0) : THuncModel(sigma,  nOrder,probHplus) {}
  THuncModel_Scale(double *sigma, double *r, int nOrder, double probHplusminus=0) : THuncModel(sigma,r,nOrder,probHplusminus), fgamma(1) {}
  THuncModel_Scale(QCDobservable *obs, double mu,        double probHplusminus=0) : THuncModel(obs,mu,        probHplusminus), fgamma(1) {}
  THuncModel_Scale(const THuncModel_Scale& other) : THuncModel(other) { fgamma = other.fgamma; }
  //
  //double P_deltanp1(double delta_np1, int n); // this is the probability distribution for delta_{n+1}
  //double Var_deltanp1();  // this is the variance of the same distribution
  double P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n);
  double P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n);
  double P_deltanp1_dots_delta1_H0(double delta_np1, int n);
  double P_deltan_dots_delta1_H0(int n);
  //
  void SetPrior(double gamma=1) { fgamma = gamma; }
  //
  double P_lambda_H0(double lambda, int n);
  //
protected:
  double fgamma;
  double Prior(double lambda);
};






// customizeable model
//
class THuncModel_Custom : public THuncModel {
private:
  double computeIntegral();
  double computeIntegralMarginal();
public:
  THuncModel_Custom(double *sigma,            int nOrder, double probHplusminus=0) : THuncModel(sigma,  nOrder,probHplusminus), fdeltaNorm(NULL), fParsNorm(NULL) {}
  THuncModel_Custom(double *sigma, double *r, int nOrder, double probHplusminus=0) : THuncModel(sigma,r,nOrder,probHplusminus), fdeltaNorm(NULL), fParsNorm(NULL) {}
  THuncModel_Custom(QCDobservable *obs, double mu,        double probHplusminus=0) : THuncModel(obs,mu,        probHplusminus), fdeltaNorm(NULL), fParsNorm(NULL) {}
  THuncModel_Custom(const THuncModel_Custom& other) : THuncModel(other) {}
  //
  void SetNumberOfParameters(int nPar) { fnPar = nPar; }
  //
  double P_deltanp3_dots_delta1_H0(double delta_np3, double delta_np2, double delta_np1, int n);
  double P_deltanp2_dots_delta1_H0(double delta_np2, double delta_np1, int n);
  double P_deltanp1_dots_delta1_H0(double delta_np1, int n);
  double P_deltan_dots_delta1_H0(int n);
  //
  double P_pars_H0(int n, const double *par);
  double P_par_marginal_H0(int n, double par, int whichpar);
  //
protected:
  int fnPar;
  friend int integrandCUBAcustom(const int *ndim, const double x[], const int *ncomp, double res[], void *pars);
  friend double integrandGaussCustom(double delta_np1, void *par);
  friend int integrandCUBAcustomNorm(const int *ndim, const double x[], const int *ncomp, double res[], void *model);
  friend int integrandCUBAcustomMarginal(const int *ndim, const double x[], const int *ncomp, double res[], void *pars);
  double norm_custom_P_deltan_given_pars(double *delta_k, int k, const double *pars);
  double norm_custom(double *delta, double *rhatk, int n, const double *pars);
  double P_deltan_dots_delta1_custom_integrand(int n, double *delk, double *rhat, const double *par);
  double P_par_marginal_integrand(int n, double *delk, double *rhat, double parFixed, int whichPar, const double *par);
  double *fdeltaNorm, *frhatkNorm, *fParsNorm, fParFixed;
  int fnNorm, fwhichPar;
};
extern double custom_P_deltan_given_pars(double *delta_k, double *rhat_k, int k, const double *par);
extern double custom_P0(const double *par);




